/*
 * local_dungeon_view.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "control.hpp"
#include "controller.hpp"
#include "draw.hpp"
#include "local_dungeon_view.hpp"
#include "mini_map_colour.hpp"
#include "platform/display.hpp"
#include "platform/sound_manager.hpp"

// NB item_depth must be greater than the other depths here.
const int icon_depth = -120;
const int entity_depth = -110;
const int item_depth = -100;

const int MY_X_OFFSCREEN_VALUE = -999;

LocalDungeonView::LocalDungeonView(const Controller & ctrlr)
	: gvt(0), current_room(-1),
	  my_x(MY_X_OFFSCREEN_VALUE), my_y(0), my_facing(D_NORTH),
	  my_approached(false), 
	  menu_gfx_centre(0), menu_gfx_empty(0),
	  controller(ctrlr), tap_control(0), menu_open(false), menu_highlight(false),
	  fire_start_time(0), attack_mode(false), menu_null(M_OK),
	  last_known_time(0), last_known_x(0), last_known_y(0),
	  last_mtime(0), flash_screen_start(-9999),
	  skull_renderer(0), potion_renderer(0),
	  nskulls(0), health(0), magic(NO_POTION), poison_immun(false),
	  won(false), lost(false),
	  cheat_controls(0)
{
	for (int i=0; i<4; ++i) menu_control[i] = 0;
}

//
// modifiers
//

void LocalDungeonView::setMiniMap(int w, int h)
{
	mini_map.create(w, h);
}

void LocalDungeonView::setCurrentRoom(int r, int w, int h)
{
	typedef map<int, RoomData>::iterator Iter;

	// Wipe all entities from the current map
	RoomData &rd(rooms[current_room]);
	rd.emap.clear();

	// Create the new map (if necessary) and reset current_room & current_dmap.
	Iter it = rooms.find(r);
	if (it == rooms.end()) {
		pair<Iter,bool> p = rooms.insert(make_pair(r, RoomData()));
		ASSERT(p.second);
		it = p.first;
		it->second.create(w, h);
	}
	current_room = r;
}

int LocalDungeonView::getCurrentRoom() const
{
	return current_room;
}

void LocalDungeonView::updateKnightIndicator()
{
	// Set knight indicator (highlight 0)

	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (my_x == MY_X_OFFSCREEN_VALUE || ri == rooms.end()
	|| ri->second.origin.isNull()) {
		// Knight not found, OR current room is not mapped
		// Don't draw a highlight.
		mini_map.setHighlight(MapCoord(), 0);
	} else {
		// Knight exists and current room is mapped
		// Draw the highlight.
		mini_map.setHighlight(MapCoord(ri->second.origin.getX()+my_x,
									   ri->second.origin.getY()+my_y), 0);
	}
}

void LocalDungeonView::mapCurrentRoom(const MapCoord &bl_pos)
{
	// Set the "origin" field in the current RoomData.
	map<int,RoomData>::iterator it = rooms.find(current_room);
	if (it == rooms.end()) return;
	it->second.origin = bl_pos;

	// Copy it to the mini-map if necessary.
	if (it->second.awaiting_copy) {
		copyRoomToMiniMap(it->second);
	}
	
	// Set the knight indicator (highlight 0).
	updateKnightIndicator();
}

void LocalDungeonView::mapWall(const MapCoord &mc)
{
	mini_map.setColour(mc, COL_WALL);
}

void LocalDungeonView::mapKnightLocation(int id, const MapCoord &pos)
{
	// We add 1 to the ID, in order to prevent clashes with ID 0 which is always the
	// current knight's location.
	mini_map.setHighlight(pos, id+1);
}

void LocalDungeonView::mapItemLocation(const MapCoord &mc, bool on)
{
	if (mc.isNull()) return;
	const int id = mc.getY() * mini_map.getHeight() + mc.getX() + 10; // add 10 to prevent clashes with other IDs
	if (on) {
		mini_map.setHighlight(mc, id);
	} else {
		mini_map.setHighlight(MapCoord(), id);
	}
}

void LocalDungeonView::wipeMap()
{
	// Wipe memorized room positions
	for (map<int, RoomData>::iterator it = rooms.begin(); it != rooms.end(); ++it) {
		it->second.origin = MapCoord();
		it->second.awaiting_copy = true;
	}

	// Wipe the mini-map itself
	mini_map.wipe();
}

bool LocalDungeonView::isSquareMapped(const MapCoord &mc) const
{
	if (mini_map.getColour(mc) == COL_UNMAPPED) return false; else return true;
}

void LocalDungeonView::copyRoomToMiniMap(RoomData &rd)
{
	if (!rd.awaiting_copy) return;
	if (rd.origin.isNull()) return;
	
	rd.awaiting_copy = false;
	const int ox = rd.origin.getX();
	const int oy = rd.origin.getY();

	for (int i=0; i<rd.width; ++i) {
		for (int j=0; j<rd.height; ++j) {
			const qlist<RoomData::GfxEntry> & gfx(rd.lookupGfx(i,j));
			MiniMapColour col = COL_FLOOR;
			for (qlist<RoomData::GfxEntry>::const_iterator it = gfx.begin(); it != gfx.end();
			++it) {
				if (it->col != COL_UNMAPPED) {
					col = it->col;
					break;
				}
			}
			mini_map.setColour(MapCoord(i+ox, j+oy), col);
		}
	}
}

void LocalDungeonView::addEntity(unsigned short int id, int x, int y, MapHeight ht, MapDirection facing,
								 const Anim * anm, const Overlay * ovr, int af, int atz_diff,
								 bool ainvuln,
								 int cur_ofs, MotionType mt, int mtr)
{
	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (ri == rooms.end()) return;
	ri->second.emap.addEntity(gvt, id, x, y, ht, facing, anm, ovr, af, atz_diff+gvt, ainvuln,
							  cur_ofs, mt, mtr);
	if (id==0) {
		my_x = x;
		my_y = y;
		if (my_x >= 0) {
			last_known_x = my_x;
			last_known_y = my_y;
		}
		my_approached = false;
		updateKnightIndicator();
	}
}

void LocalDungeonView::rmEntity(unsigned short int id)
{
	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (ri == rooms.end()) return;
	ri->second.emap.rmEntity(id);
	if (id==0) {
		my_x = MY_X_OFFSCREEN_VALUE;
		updateKnightIndicator();
	}
}

void LocalDungeonView::repositionEntity(unsigned short int id, int new_x, int new_y)
{
	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (ri == rooms.end()) return;
	ri->second.emap.repositionEntity(id, new_x, new_y);
	if (id == 0) {
		my_x = new_x;
		my_y = new_y;
		if (my_x >= 0) {
			last_known_x = my_x;
			last_known_y = my_y;
		}
		my_approached = false;
		updateKnightIndicator();
	}
}

void LocalDungeonView::moveEntity(unsigned short int id, MotionType mt, int motion_dur, bool missile_mode)
{
	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (ri == rooms.end()) return;
	ri->second.emap.moveEntity(gvt, id, mt, motion_dur, missile_mode);
	if (id==0) {
		my_approached = (mt == MT_APPROACH);
		updateKnightIndicator();
	}
}

void LocalDungeonView::flipEntityMotion(unsigned short int id, int initial_delay, int motion_dur)
{
	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (ri == rooms.end()) return;
	ri->second.emap.flipEntityMotion(gvt, id, initial_delay, motion_dur);
	if (id == 0) {
		// not sure if updateKnightIndicator is strictly necessary here,
		// but we'll have it anyway
		updateKnightIndicator();
	}
}	

void LocalDungeonView::setAnimData(unsigned short int id, const Anim *a, const Overlay *o,
								   int af, int atz_diff, bool ainvuln, bool currently_moving)
{
	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (ri == rooms.end()) return;
	ri->second.emap.setAnimData(id, a, o, af, atz_diff ? atz_diff+gvt : 0, ainvuln, currently_moving);
}

void LocalDungeonView::setFacing(unsigned short int id, MapDirection new_facing)
{
	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (ri == rooms.end()) return;	
	ri->second.emap.setFacing(id, new_facing);
	if (id==0) {
		updateKnightIndicator();
	}
}

void LocalDungeonView::clearTiles(int x, int y)
{
	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (ri == rooms.end()) return;
	if (!ri->second.valid(x,y)) return;
	qlist<RoomData::GfxEntry> &gfx(ri->second.lookupGfx(x, y));
	// We know that the tiles come before items or other stuff like that.
	while (!gfx.empty() && gfx.front().depth != item_depth) gfx.pop_front();
}

void LocalDungeonView::setTile(int x, int y, int depth, const Graphic * gfx,
							   shared_ptr<const ColourChange> cc, MiniMapColour col)
{
	map<int,RoomData>::iterator ri = rooms.find(current_room);
	if (ri == rooms.end()) return;
	if (!ri->second.valid(x,y)) return;
	qlist<RoomData::GfxEntry> &glst(ri->second.lookupGfx(x, y));
	qlist<RoomData::GfxEntry>::iterator it = glst.begin();
	while (1) {
		// The list is in sorted order, highest depth first.
		if (it == glst.end() || depth > it->depth) {
			if (gfx) {
				RoomData::GfxEntry ge;
				ge.gfx = gfx;
				ge.cc = cc;
				ge.col = col;
				ge.depth = depth;
				glst.insert(it, ge);
			}
			break;
		} else if (it->depth == depth) {
			if (gfx) {
				it->gfx = gfx;
				it->cc = cc;
				it->col = col;
			} else {
				glst.erase(it);
			}
			break;
		} else {
			++it;
		}
	}
	
	ri->second.awaiting_copy = true;
}

void LocalDungeonView::setItem(int x, int y, const Graphic * graphic)
{
	// We can implement this using setTile.
	setTile(x, y, item_depth, graphic, shared_ptr<const ColourChange>(), COL_UNMAPPED);
}


void LocalDungeonView::placeIcon(int x, int y, const Graphic *g, int dur)
{
	if (current_room == -1) return;
	
	// Add the 'icon' to the gmap. Also store in 'icons' list.
	setTile(x, y, icon_depth, g, shared_ptr<const ColourChange>(), COL_UNMAPPED);
	LocalIcon ic;
	ic.room_no = current_room;
	ic.x = x;
	ic.y = y;
	ic.expiry = dur + gvt;
	icons.push(ic);
}

void LocalDungeonView::setBackpack(int slot, const Graphic *gfx, int no_carried, int no_max)
{
	BackpackEntry be;
	be.gfx = gfx;
	be.no_carried = no_carried;
	be.no_max = no_max;
	backpack.erase(slot);
	backpack.insert(make_pair(slot, be));
}

void LocalDungeonView::addSkull()
{
	++nskulls;
}

void LocalDungeonView::setHealth(int h)
{
	health = h;
}

void LocalDungeonView::setPotionMagic(PotionMagic pm, bool pi)
{
	magic = pm;
	poison_immun = pi;
}

void LocalDungeonView::flashMessage(const string &s)
{
	Message m;
	m.message = s;
	m.time = gvt;
	messages.push_back(m);
}

void LocalDungeonView::cancelContinuousMessages()
{
	cts_messages.clear();
}

void LocalDungeonView::addContinuousMessage(const string &s)
{
	cts_messages.push_back(s);
}

void LocalDungeonView::flashScreen(int delay)
{
	flash_screen_start = gvt + delay;
}

void LocalDungeonView::playSound(const string &name, int frequency)
{
	SoundManager::instance().playSound(name, frequency);
}


//
// Menus/Controls
//

void LocalDungeonView::setMenuControl(MapDirection d, const Control *ctrl,
									  const Control *prev)
{
	if (menu_null==M_CTS && menu_null_dir==d && ctrl != prev) {
		// This stops e.g. immediately opening a door after the lock picks completed.
		menu_null = M_NULL;
	}
	menu_control[d] = ctrl;
}

void LocalDungeonView::setAvailableControls(const vector<pair<const Control *,bool> >
											&controls)
{
	const Control * prev_menu_control[4];
	for (int i=0; i<4; ++i) {
		prev_menu_control[i] = menu_control[i];
		menu_control[i] = 0;
	}
	tap_control = 0;
	int tap_pri = 0;

	// the server is currently telling us that all controls in "controls" are currently
	// possible and available to our knight. so put those controls into the menu or
	// "tap_control" as appropriate.

	for (vector<pair<const Control *, bool> >::const_iterator ctrl = controls.begin(); 
	ctrl != controls.end(); ++ctrl) {
		// Find the (primary) control with the highest tap priority:
		// (NB: control is primary iff ctrl->second == true.)
		if (ctrl->first->getTapPriority() > tap_pri && ctrl->second) {
			tap_pri = ctrl->first->getTapPriority();
			tap_control = ctrl->first;
		}
		if (ctrl->first->getMenuGraphic() != 0) {
			MapDirection d = ctrl->first->getMenuDirection();
			// If no control in direction d, or if the control in direction d has
			// property MS_WEAK, then save ctrl->first into menu position d.
			if (menu_control[d] == 0 ||
			(menu_control[d]->getMenuSpecial() & Control::MS_WEAK)) {
				setMenuControl(d, ctrl->first, prev_menu_control[d]);
			}
		}
	}

	// Now try to place controls that could not be put in their preferred position.
	for (vector<pair<const Control *, bool> >::const_iterator ctrl = controls.begin(); 
	ctrl != controls.end(); ++ctrl) {
		if (ctrl->first->getMenuGraphic() != 0) {
			MapDirection d = ctrl->first->getMenuDirection();
			if (menu_control[d] != ctrl->first &&
			(ctrl->first->getMenuSpecial() & Control::MS_WEAK) == 0) {
				for (int i=0; i<4; ++i) {
					if (menu_control[i] == 0) {
						setMenuControl(MapDirection(i), ctrl->first, prev_menu_control[i]);
						break;
					}
				}
			}
		}
	}
}

void LocalDungeonView::setMenuHighlight(const Control *ctrl)
{
	if (ctrl) {
		for (int i=0; i<4; ++i) {
			if (menu_control[i] == ctrl) {
				menu_highlight = true;
				menu_highlight_dir = MapDirection(i);
				return;
			}
		}
	}
	menu_highlight = false;
}

const Control * LocalDungeonView::readController()
{
	// This routine reads the joystick/controller, and converts this
	// information to an appropriate "Control". It also displays the
	// on-screen menu (when fire is held down) if necessary.

	// set up some constants
	const int menu_delay = CfgInt("menu_delay"); // cutoff btwn 'tapping' and 'holding' fire.
	const bool approached = my_approached;

	// get current controller state
	MapDirection ctrlr_dir;
	bool ctrlr_centred, ctrlr_fire, suicide_keys;
	controller.get(ctrlr_dir, ctrlr_centred, ctrlr_fire, suicide_keys);

	// if suicide keys held, then we can exit immediately...
	if (suicide_keys) return g_standard_controls[SC_SUICIDE];

	// check the cheat keys:
	int k = controller.getCheatKey();
	if (cheat_controls && k >= 1 && k <= cheat_controls->size()) {
		return (*cheat_controls)[k-1];
	}
	
	// work out 'long' fire state
	enum FireState { NONE, TAPPED, HELD } long_fire = NONE;
	if (ctrlr_fire) {
		if (fire_start_time == 0) {
			// fire_start_time: set to time when fire was pressed down (or 0 if fire is
			// currently released)
			// attack_mode: if fire+direction is used to attack, then attack_mode is set true,
			// which disables menus and 'tapping'.
			fire_start_time = gvt;
			attack_mode = false;
		}
		if (gvt >= fire_start_time + menu_delay && !attack_mode) {
			long_fire = HELD;
		}
	} else {
		// do a 'tap' if allowed
		if (fire_start_time > 0 && fire_start_time < gvt + menu_delay && !attack_mode) {
			long_fire = TAPPED;
		}
		// reset fire state when fire is released
		fire_start_time = 0;
		attack_mode = false;
	}

	
	if (menu_open) {
		// Menu open

		if (long_fire != HELD) {
			// close the menu
			menu_open = false;
			menu_null = M_OK;
		} else {
			if (menuNull() && (ctrlr_centred || ctrlr_dir != menu_null_dir)) {
				// Cancel null dir.
				menu_null = M_OK;
			}
			if (!ctrlr_centred && (!menuNull() || ctrlr_dir != menu_null_dir)) {
				// A menu option has been selected
				const Control *c = menu_control[ctrlr_dir];
				if (c) {
					// stop the action being executed more than once in a row.
					// (continuous actions will be stopped only if the control disappears,
					// which is what M_CTS is for.)
					menu_null = (c->isContinuous()? M_CTS: M_NULL);
					menu_null_dir = ctrlr_dir;
				}
				return c;
			}
		}

	} else if (approached) {
		// Approached, and menu closed
		if ((ctrlr_centred || ctrlr_dir != my_facing) && !ctrlr_fire) {
			// Withdraw if controller is no longer pointing in the approach direction
			// and fire is released.
			return g_standard_controls[SC_WITHDRAW];
		} else {
			if (long_fire == HELD) {
				menu_open = true;
				menu_null = M_NULL;
				menu_null_dir = my_facing;
			} else if (long_fire == TAPPED) { 
				return tap_control;
			}
		}
	} else {
		// Not approached, and menu closed
		if (ctrlr_centred) {
			if (long_fire == HELD) {
				menu_open = true;
				menu_null = M_OK;
			} else if (long_fire == TAPPED) { 
				return tap_control;
			}
		} else {
			if (ctrlr_fire) {
				attack_mode = true;
				return g_standard_controls[SC_ATTACK + ctrlr_dir];
			} else {
				my_facing = ctrlr_dir;
				return g_standard_controls[SC_MOVE + ctrlr_dir];
			}
		}
	}

	// No commands coming from the controller at present.
	return 0;
}



//
// drawing
//

#ifdef _WIN32_WCE
namespace {
	void BlendPixel(Display &disp, int x, int y)
	{
		int r, g, b;
		disp.readPixel(x,y,r,g,b);
		r = r * 2 / 3;
		g = g * 2 / 3;
		b = b * 2 / 3 + (255/3);
		disp.drawPixel(x,y,r,g,b);
	}
}
#endif

void LocalDungeonView::draw(Display &disp, int left, int bottom, int new_gvt)
{
	gvt = new_gvt;
	
	// (TODO) perhaps these config ints should be cached at construction time??
	const int dungeon_left = CfgInt("dpy_dungeon_left") + left;
	const int dungeon_bottom = CfgInt("dpy_dungeon_bottom") + bottom;
	const int dungeon_right = CfgInt("dpy_dungeon_right") + left;
	const int dungeon_top = CfgInt("dpy_dungeon_top") + bottom;
	const int inventory_left = CfgInt("dpy_inventory_left") + left;
	const int inventory_bottom = CfgInt("dpy_inventory_bottom") + bottom;
	const int inventory_right = CfgInt("dpy_inventory_right") + left;
	const int inventory_top = CfgInt("dpy_inventory_top") + bottom;
	const int mini_map_left = CfgInt("dpy_map_left") + left;
	const int mini_map_bottom = CfgInt("dpy_map_bottom") + bottom;
	const int skulls_left = CfgInt("dpy_skulls_left") + left;
	const int skulls_bottom = CfgInt("dpy_skulls_bottom") + bottom;
	const int potion_left = CfgInt("dpy_potion_left") + left;
	const int potion_bottom = CfgInt("dpy_potion_bottom") + bottom;
	const int pixels_per_square = CfgInt("pixels_per_square");
	const int mm_scale = CfgInt("mini_map_scale");

	// simple calculation of screen extents (for screen flash)
	const int scrn_left = min(dungeon_left, inventory_left);
	const int scrn_right = max(dungeon_right, inventory_right);
	const int scrn_bottom = min(dungeon_bottom, inventory_bottom);
	const int scrn_top = max(dungeon_top, inventory_top);
	
	// If game is over, we do something different...
	if (won || lost) {
		if (won) {
			disp.drawText(dungeon_left,100,"WINNER"); // (TODO)
		} else {
			disp.drawText(dungeon_left,100,"LOSER"); // (TODO)
		}
		return;
	}

	// Clear out expired icons
	while (!icons.empty() && gvt > icons.top().expiry) {
		map<int,RoomData>::iterator ri = rooms.find(icons.top().room_no);
		if (ri != rooms.end()) {
			qlist<RoomData::GfxEntry> &glst(ri->second.lookupGfx(icons.top().x,
																 icons.top().y));
			for (qlist<RoomData::GfxEntry>::iterator it = glst.begin(); it != glst.end();
			++it) {
				if (it->depth == icon_depth) {
					glst.erase(it);
					break;
				}
			}
		}
		icons.pop();
	}

	// Clear out expired messages
	while (!messages.empty() && gvt >= messages.front().time + CfgInt("message_total_time")) {
		messages.pop_front();
	}

	// Check whether screen should flash
	const bool screen_flash = (gvt >= flash_screen_start && gvt < flash_screen_start + CfgInt("screen_flash_time"));

	// Draw the coloured background if screen is flashing
	if (screen_flash)
	{
		// Rather crude method!
		const int r = CfgInt("screen_flash_r");
		const int g = CfgInt("screen_flash_g");
		const int b = CfgInt("screen_flash_b");
		disp.beginPixels();
		for (int y = scrn_bottom; y < scrn_top; ++y) {
			for (int x = scrn_left; x < scrn_right; ++x) {
				disp.drawPixel(x, y, r, g, b);
			}
		}
		disp.endPixels();
	}

	// Draw current room
	if (my_x != MY_X_OFFSCREEN_VALUE) last_known_time = gvt;
	if (gvt < last_known_time + CfgInt("death_draw_map_time")) {
		// We are alive, or have been dead for no longer than "death_draw_map_time".
		// Draw the map.
		map<int,RoomData>::iterator ri = rooms.find(current_room);
		if (ri != rooms.end()) {
			RoomData &rd(ri->second);

			// Clear out the graphics buffer
			for (map<int, vector<GraphicElement> >::iterator it = gfx_buffer.begin();
			it != gfx_buffer.end(); ++it) {
				it->second.clear();
			}

			// Set clip rectangle
			disp.setClipRectangle(dungeon_left, dungeon_bottom, dungeon_right,
								  dungeon_top);

			// Find on-screen bottom-left corner for the room
			const int room_bl_x = (dungeon_right - dungeon_left - rd.width*pixels_per_square)
				/ 2 + dungeon_left;
			const int room_bl_y = (dungeon_top - dungeon_bottom - rd.height*pixels_per_square)
				/ 2 + dungeon_bottom;

			// Build a list of graphics to be drawn

			// Draw tiles (if screen not flashing)
			if (!screen_flash) {
				for (int i=0; i<rd.width; ++i) {
					for (int j=0; j<rd.height; ++j) {
						const int screen_x = i*pixels_per_square + room_bl_x;
						const int screen_y = j*pixels_per_square + room_bl_y;
						
						const qlist<RoomData::GfxEntry> &glst(rd.lookupGfx(i,j));
						for (qlist<RoomData::GfxEntry>::const_iterator it = glst.begin();
							 it != glst.end(); ++it) {
							GraphicElement ge;
							ge.sx = screen_x;
							ge.sy = screen_y;
							ge.gr = it->gfx;
							ge.cc = it->cc.get();
							gfx_buffer[it->depth].push_back(ge);
						}
					}
				}
			}

			// Add the entities (always)
			rd.emap.getEntityGfx(gvt, room_bl_x, room_bl_y, entity_depth, gfx_buffer);

			// Now draw all buffered graphics
			for (map<int, vector<GraphicElement> >::reverse_iterator it = gfx_buffer.rbegin();
			it != gfx_buffer.rend(); ++it) {
				for (vector<GraphicElement>::const_iterator it2 = it->second.begin();
				it2 != it->second.end(); ++it2) {
					if (it2->cc) {
						disp.drawGraphicWithCC(it2->sx, it2->sy, *it2->gr, *it2->cc);
					} else {
						disp.drawGraphic(it2->sx, it2->sy, *it2->gr, false);
					}
				}
			}
		}
	}

	// The rest does not need to be drawn when screen is flashing

	if (!screen_flash) {
		
		// Draw inventory icons. (This is handled by DrawUI)
		for (map<int, BackpackEntry>::iterator it = backpack.begin(); it != backpack.end();
			 ++it) {
			DrawUI::drawBackpackEntry(disp, inventory_left, inventory_bottom,
									  inventory_right, inventory_top,
									  it->first, it->second.gfx, it->second.no_carried,
									  it->second.no_max);
		}
	
		// Draw Mini-Map or Menu
		if (menu_open) {
			// Menu is drawn in same position as mini-map.
			DrawUI::drawMenu(disp, mini_map_left, mini_map_bottom,
							 menu_gfx_centre, menu_gfx_empty, menu_gfx_highlight,
							 menu_control[D_NORTH]? menu_control[D_NORTH]->getMenuGraphic() : 0,
							 menu_control[D_EAST]? menu_control[D_EAST]->getMenuGraphic() : 0,
							 menu_control[D_SOUTH]? menu_control[D_SOUTH]->getMenuGraphic() : 0,
							 menu_control[D_WEST]? menu_control[D_WEST]->getMenuGraphic() : 0,
							 menu_highlight, menu_highlight_dir);
		} else {
			// Mini-Map is only drawn when menu is closed
			if (mm_scale == 0) {
				mini_map.drawOnePointFiveScale(disp, mini_map_left, mini_map_bottom, gvt);
			} else {
				mini_map.draw(disp, mini_map_left, mini_map_bottom, mm_scale, gvt);
			}
		}

#ifdef _WIN32_WCE
		{
			// Draw a blob at the mouse_centre position.
			disp.beginPixels();
			static const int x0 = CfgInt("mouse_centre_x");
			static const int y0 = CfgInt("mouse_centre_y");
			for (int i=-3; i<=3; ++i) {
				for (int j=-3; j<=3; ++j) {
					if ((i!=-3 && i!=3) || (j!=-3 && j!=3)) {
						BlendPixel(disp, x0+i, y0+j);
					}
				}
			}
			disp.endPixels();
		}
#endif		
			
		// Draw Skulls
		if (skull_renderer) {
			skull_renderer->draw(disp, nskulls, skulls_left, skulls_bottom);
		}

		// Draw Potion Bottle
		if (potion_renderer) {
			potion_renderer->draw(gvt, disp, health, magic, poison_immun,
								  potion_left, potion_bottom);
		}
	
		// Determine whether we should draw a message
		if (messages.empty() && cts_messages.empty()) {
			last_mtime = 0;
		} else {
			if (last_mtime == 0) {
				if (!messages.empty()) last_mtime = messages.front().time;
				else last_mtime = gvt;
			}
			int mtime = gvt - last_mtime;

			const int nmsgs = messages.size() + cts_messages.size();
			int mphase = (mtime / (CfgInt("message_flash_time") * 2)) % nmsgs;
			const string &the_msg(mphase < messages.size() ? messages[mphase].message
								  : cts_messages[mphase - messages.size()]);
			mphase = (mtime / CfgInt("message_flash_time")) % 2;
			if (mphase == 0) {
				map<int,RoomData>::iterator ri = rooms.find(current_room);
				if (ri != rooms.end()) {
					DrawUI::drawMessage(disp, dungeon_left, dungeon_bottom, dungeon_right,
										dungeon_top, last_known_x, last_known_y, ri->second.width,
										ri->second.height, the_msg);
				}
			}
		}

	}
}

