/*
 * colour_change.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "binary_io/binary_io.hpp"
#include "colour_change.hpp"

#include <algorithm>
using namespace std;

void ColourChange::add(int ro, int go, int bo, int rn, int gn, int bn)
{
	RGB o, n;
	o.r = ro; o.g = go; o.b = bo;
	n.r = rn; n.g = gn; n.b = bn;

	// NB would be faster to do some kind of incremental sort (eg an insertion sort),
	// but it's easier just to append the new value and then re-sort the whole lot :)
	mappings.push_back(make_pair(o,n));
	sort(mappings.begin(), mappings.end());
}

class CmpFirstMasked {
public:
	explicit CmpFirstMasked(int mask_) : mask(mask_) { }
	bool operator()(const pair<RGB,RGB> &lhs, const pair<RGB,RGB> &rhs) {
		const int r1 = (lhs.first.r & mask) - (rhs.first.r & mask);
		if (r1 < 0) return true;
		if (r1 > 0) return false;
		const int g1 = (lhs.first.g & mask) - (rhs.first.g & mask);
		if (g1 < 0) return true;
		if (g1 > 0) return true;
		const int b1 = (lhs.first.b & mask) - (rhs.first.b & mask);
		if (b1 < 0) return true;
		return false;
	}
private:
	int mask;
};

bool ColourChange::lookup(int ro, int go, int bo, int &rn, int &gn, int &bn) const
{
	pair<RGB,RGB> x;
	x.first.r = ro;
	x.first.g = go;
	x.first.b = bo;

	CmpFirstMasked cmp(MASK);
	vector<pair<RGB,RGB> >::const_iterator it = lower_bound(mappings.begin(),
		mappings.end(), x, cmp);
	if (it == mappings.end() || cmp(x, *it)) {
		return false;
	} else {
		rn = it->second.r;
		gn = it->second.g;
		bn = it->second.b;
		return true;
	}
}

//
// serialization
//

void ColourChange::serialize(ostream &str) const
{
	BinWriteInt(str, mappings.size());
	for (int i=0; i<mappings.size(); ++i) {
		BinWriteUbyte(str, mappings[i].first.r);
		BinWriteUbyte(str, mappings[i].first.g);
		BinWriteUbyte(str, mappings[i].first.b);
		BinWriteUbyte(str, mappings[i].second.r);
		BinWriteUbyte(str, mappings[i].second.g);
		BinWriteUbyte(str, mappings[i].second.b);
	}
}

ColourChange::ColourChange(istream &str)
{
	mappings.resize(BinReadInt(str));
	for (int i=0; i<mappings.size(); ++i) {
		mappings[i].first.r = BinReadUbyte(str);
		mappings[i].first.g = BinReadUbyte(str);
		mappings[i].first.b = BinReadUbyte(str);
		mappings[i].second.r = BinReadUbyte(str);
		mappings[i].second.g = BinReadUbyte(str);
		mappings[i].second.b = BinReadUbyte(str);
	}
}


